#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import itertools

import titus.P as P
from titus.datatype import Type
from titus.datatype import FcnType
from titus.datatype import ExceptionType
from titus.datatype import AvroType
from titus.datatype import AvroCompiled
from titus.datatype import AvroNull
from titus.datatype import AvroBoolean
from titus.datatype import AvroInt
from titus.datatype import AvroLong
from titus.datatype import AvroFloat
from titus.datatype import AvroDouble
from titus.datatype import AvroBytes
from titus.datatype import AvroFixed
from titus.datatype import AvroString
from titus.datatype import AvroEnum
from titus.datatype import AvroArray
from titus.datatype import AvroMap
from titus.datatype import AvroRecord
from titus.datatype import AvroField
from titus.datatype import AvroUnion

class IncompatibleTypes(Exception):
    """An internal Exception raised when signature matching needs to give up and return False.

    This is an internal Exception and should always be caught, not propagated to the user.
    """
    pass

class PFAVersion(object):
    """Describes a PFA (the language) version number as a triple of major number, minor number, release number (all non-negative integers).

    The version has a strict ordering (numerical lexicographic).

    Used to specify a version in which to interpret a PFA document and to label the beginning and end of support of function signatures.

    It could, in the future, be used to label support ranges of special forms, too, but that would always be managed by hand within the special form's constructors.
    """

    def __init__(self, major, minor, release):
        """Create a ``PFAVersion`` from a major, minor, release number triple (all non-negative integers).

        :type major: non-negative integer
        :param major: major version number (major changes)
        :type minor: non-negative integer
        :param minor: minor version number (new features)
        :type release: non-negative integer
        :param release: release version number (bug-fixes)
        """
        if major < 0 or minor < 0 or release < 0 or not isinstance(major, int) or not isinstance(minor, int) or not isinstance(release, int):
            raise ValueError("PFA version major, minor, and release numbers must be non-negative integers")
        self.major = major
        self.minor = minor
        self.release = release
    def __repr__(self):
        """Represent a PFAVersion as a dotted triple on the screen."""
        return "{0}.{1}.{2}".format(self.major, self.minor, self.release)
    def __lt__(self, other):
        """Strict ordering on PFAVersions: check major number first, then minor, then release."""
        if self.major == other.major and self.minor == other.minor:
            return (self.release < other.release)
        elif self.major == other.major:
            return (self.minor < other.minor)
        else:
            return (self.major < other.major)
    def __ge__(self, other):
        """Strict ordering on PFAVersions: check major number first, then minor, then release."""
        if self.major == other.major and self.minor == other.minor:
            return (self.release >= other.release)
        elif self.major == other.major:
            return (self.minor >= other.minor)
        else:
            return (self.major >= other.major)     
    def __cmp__(self, other):
        """Strict ordering on PFAVersions: check major number first, then minor, then release."""
        if self.major == other.major and self.minor == other.minor:
            return ((self.release > other.release) - (self.release < other.release))
        elif self.major == other.major:
            return ((self.minor > other.minor) - (self.minor < other.minor))
        else:
            return ((self.major > other.major) - (self.major < other.major))
    @staticmethod
    def fromString(x):
        """Create a ``PFAVersion`` from a dotted string.

        :type x: string in the form "major.minor.release"
        :param x: string to convert into a ``PFAVersion``
        :rtype: titus.signature.PFAVersion
        :return: corresponding ``PFAVersion``
        """
        try:
            major, minor, release = list(map(int, x.split(".")))
            return PFAVersion(major, minor, release)
        except ValueError:
            raise ValueError("PFA version numbers must have major.minor.release structure, where major, minor, and release are integers")

class Lifespan(object):
    """Describes the range of support of a function signature (or, in the future, special form) in terms of an optional beginning of life (birth), and optional deprecation and end of life (death).

    If a deprecation is specified, a death must be as well, and vice-versa. Whether or not a birth is specified is independent.

    At a given titus.signature.PFAVersion, the Lifespan has three possible states: current (method ``current`` returns ``True``), deprecated (method ``deprecated`` returns ``True``), and non-existent (both ``current`` and ``deprecated`` return ``False``). Method ``current`` and ``deprecated`` are mutually exclusive; for a given titus.signature.PFAVersion, they would never both return ``True``.
    """

    def __init__(self, birth, deprecation, death, contingency):
        """Create a Lifespan from a birth, deprecation, death, and contingency.

        :type birth: titus.signature.PFAVersion or ``None``
        :param birth: first PFA version number in which this signature (or special form) is valid; ``None`` means the beginning of time
        :type deprecation: titus.signature.PFAVersion or ``None``
        :param deprecation: first PFA version number in which this signature should raise a deprecation warning; ``None`` means never
        :type death: titus.signature.PFAVersion or ``None``
        :param death: first PFA version number in which this signature is skipped over as not existing; ``None`` means the end of time
        :type contingency: string or ``None``
        :param contingency: message to accompany the deprecation warning; usually tells the user which function to use instead
        """

        if deprecation is None and death is None:
            pass
        elif deprecation is not None and death is not None and deprecation < death:
            pass
        else:
            raise ValueError("deprecation and death must be specified together, and deprecation version must be strictly earlier than death version")

        if birth is not None and deprecation is not None and birth >= deprecation:
            raise ValueError("if birth and deprecation are specified together, birth version must be strictly earlier than deprecation version")

        self.birth = birth
        self.deprecation = deprecation
        self.death = death
        self.contingency = contingency

    def current(self, now):
        """:type now: titus.signature.PFAVersion
        :param now: the version number to query
        :rtype: bool
        :return: ``True`` if the feature exists and is not deprecated in version ``now``, ``False`` otherwise.
        """
        return (self.birth is None or now >= self.birth) and (self.deprecation is None or now < self.deprecation)

    def deprecated(self, now):
        """:type now: titus.signature.PFAVersion
        :param now: the version number to query
        :rtype: bool
        :return: ``True`` if the feature exists and is deprecated in version ``now``, ``False`` otherwise.
        """
        return self.deprecation is not None and self.deprecation <= now < self.death

class LabelData(object):
    """Used internally to carry information about a label in a wildcard and how well it matches a prospective argument type."""
    @staticmethod
    def distinctTypes(candidates, out = None):
        if out is None:
            out = []
        for candidate in candidates:
            if isinstance(candidate, AvroUnion):
                out = LabelData.distinctTypes(candidate.types, out)
            elif any(y.accepts(candidate) for y in out):
                pass
            else:
                i = 0
                for y in out:
                    if candidate.accepts(y): break
                    i += 1
                if i == len(out):
                    out.append(candidate)
                else:
                    out[i] = candidate
        return out

    @staticmethod
    def broadestType(candidates):
        """Compute the narrowest possible supertype of a set of types.

        :type candidates: list of titus.datatype.AvroType
        :param candidates: set of types for which to find the narrowest possible supertype
        :rtype: titus.datatype.AvroType
        :return: narrowest possible supertype, usually a union of the candidates
        """

        realCandidates = [x for x in candidates if not isinstance(x, ExceptionType)]

        if len(candidates) == 0:
            return ValueError("empty list of types")
        elif len(realCandidates) == 0:
            return ValueError("list of types consists only of exception type")

        elif all(isinstance(x, AvroNull) for x in realCandidates):
            return realCandidates[0]
        elif all(isinstance(x, AvroBoolean) for x in realCandidates):
            return realCandidates[0]

        elif all(isinstance(x, AvroInt) for x in realCandidates):
            return AvroInt()
        elif all(isinstance(x, AvroInt) or isinstance(x, AvroLong) for x in realCandidates):
            return AvroLong()
        elif all(isinstance(x, AvroInt) or isinstance(x, AvroLong) or isinstance(x, AvroFloat) for x in realCandidates):
            return AvroFloat()
        elif all(isinstance(x, AvroInt) or isinstance(x, AvroLong) or isinstance(x, AvroFloat) or isinstance(x, AvroDouble) for x in realCandidates):
            return AvroDouble()

        elif all(isinstance(x, AvroBytes) for x in realCandidates):
            return realCandidates[0]
        elif all(isinstance(x, AvroString) for x in realCandidates):
            return realCandidates[0]

        elif all(isinstance(x, AvroArray) for x in realCandidates):
            return AvroArray(P.mustBeAvro(LabelData.broadestType([x.items for x in realCandidates])))

        elif all(isinstance(x, AvroMap) for x in realCandidates):
            return AvroMap(P.mustBeAvro(LabelData.broadestType([x.values for x in realCandidates])))

        elif all(isinstance(x, AvroFixed) for x in realCandidates):
            fullName = realCandidates[0].fullName
            if all(x.fullName == fullName for x in realCandidates[1:]):
                return realCandidates[0]
            else:
                raise IncompatibleTypes("incompatible fixed types: " + " ".join(map(repr, realCandidates)))

        elif all(isinstance(x, AvroEnum) for x in realCandidates):
            fullName = realCandidates[0].fullName
            if all(x.fullName == fullName for x in realCandidates[1:]):
                return realCandidates[0]
            else:
                raise IncompatibleTypes("incompatible enum types: " + " ".join(map(repr, realCandidates)))

        elif all(isinstance(x, AvroRecord) for x in realCandidates):
            fullName = realCandidates[0].fullName
            if all(x.fullName == fullName for x in realCandidates[1:]):
                return realCandidates[0]
            else:
                raise IncompatibleTypes("incompatible record types: " + " ".join(map(repr, realCandidates)))

        elif all(isinstance(x, FcnType) for x in realCandidates):
            params = realCandidates[0].params
            ret = realCandidates[0].ret

            if all(x.params == params and x.ret == ret for x in realCandidates[1:]):
                return realCandidates[0]
            else:
                raise IncompatibleTypes("incompatible function types: " + " ".join(map(repr, realCandidates)))

        elif not any(isinstance(x, FcnType) for x in realCandidates):
            types = LabelData.distinctTypes(realCandidates)
            types = [P.mustBeAvro(x) for x in types]

            countFixed = 0
            countEnum = 0
            for t in types:
                if isinstance(t, AvroFixed):
                    countFixed += 1
                if isinstance(t, AvroEnum):
                    countEnum += 1

            if countFixed > 1:
                raise IncompatibleTypes("incompatible fixed types")
            if countEnum > 1:
                raise IncompatibleTypes("incompatible enum types")

            return AvroUnion(types)

        else:
            raise IncompatibleTypes("incompatible function/non-function types: " + " ".join(map(repr, realCandidates)))

    def __init__(self):
        self.members = []
        self.expectedFields = None

    def add(self, t):
        self.members.append(t)

    def determineAssignment(self):
        out = LabelData.broadestType(self.members)
        if self.expectedFields is not None:
            if isinstance(out, AvroRecord):
                if self.expectedFields != tuple(x.name for x in out.fields):
                    raise IncompatibleTypes("fields do not agree with EnumFields symbols (same names, same order)")
            else:
                raise IncompatibleTypes("EnumFields.wildRecord does not point to a record")
        return out

    def __repr__(self):
        return "<titus.signature.LabelData {0} at {1}>".format(self.members, "0x%x" % id(self))

class Signature(object):
    """Abstract trait for function signatures. Known subclasses: titus.signature.Sig and titus.signature.Sigs."""
    def accepts(self, args, version):
        raise NotImplementedError

class Sigs(Signature):
    """PFA function signature for ad-hoc polymorphism (list of different signatures supported by the function)."""

    def __init__(self, cases):
        """:type cases: list of titus.signature.Sig
        :param cases: signatures supported by this function; the order of this list is the order in which Titus will attempt to resolve the signatures (the first one that matches is used)
        """
        self.cases = cases

    def accepts(self, args, version):
        """Determine if this list of signatures accepts the given arguments for a given PFA version number.

        :type args: list of titus.datatype.AvroType
        :param args: arguments to match against each of the signature patterns
        :type version: titus.signature.PFAVersion
        :param version: PFA version number in which to interpret the patterns
        :rtype: (titus.signature.Sig, list of titus.datatype.AvroType, AvroType)
        :return: (matching signature, resolved argument types, resolved return type) if one of the signatures accepts the arguments; ``None`` otherwise
        """
        for case in self.cases:
            result = case.accepts(args, version)
            if result is not None:
                return result
        return None

class Sig(Signature):
    """PFA function signature for a single pattern.

    This class can either be used directly as a function's only signature or it may be contained within a titus.signature.Sigs.
    """

    def __init__(self, params, ret, lifespan = Lifespan(None, None, None, None)):
        """:type params: list of titus.P
        :param params: patterns for each argument (in order)
        :type ret: titus.P
        :param ret: pattern for the return type
        :type lifespan: titus.signature.Lifespan
        :param lifespan: validity range for this signature; default is eternal (existing from the beginning of time to the end of time)
        """
        self.params = params
        self.ret = ret
        self.lifespan = lifespan

    def __repr__(self):
        alreadyLabeled = set()
        return "(" + ", ".join(list(p.keys())[0] + ": " + toText(list(p.values())[0], alreadyLabeled) for p in self.params) + " -> " + toText(self.ret, alreadyLabeled) + ")"

    def accepts(self, args, version):
        """Determine if this signature accepts the given arguments for a given PFA version number.

        :type args: list of titus.datatype.AvroType
        :param args: arguments to match against the signature pattern
        :type version: titus.signature.PFAVersion
        :param version: PFA version number in which to interpret the pattern
        :rtype: (titus.signature.Sig, list of titus.datatype.AvroType, AvroType)
        :return: (self, resolved argument types, resolved return type) if this signature accepts the arguments; ``None`` otherwise
        """

        if self.lifespan.current(version) or self.lifespan.deprecated(version):
            labelData = {}
            if len(self.params) == len(args) and all(self.check(list(p.values())[0], a, labelData, False, False) for p, a in zip(self.params, args)):
                try:
                    assignments = dict((l, ld.determineAssignment()) for l, ld in list(labelData.items()))
                    assignedParams = [self.assign(list(p.values())[0], a, assignments) for p, a in zip(self.params, args)]
                    assignedRet = self.assignRet(self.ret, assignments)
                    return (self, assignedParams, assignedRet)
                except IncompatibleTypes:
                    return None
            else:
                return None
        else:
            return None

    def check(self, pat, arg, labelData, strict, reversed):
        """Determine if a single slot in the parameter pattern matches a single argument's type.

        :type pat: titus.P
        :param pat: pattern for a type that can include wildcards, etc.
        :type arg: titus.datatype.AvroType
        :param arg: supplied argument type that may or may not fit the pattern
        :type labelData: dict from label letters to titus.signature.LabelData
        :param labelData: label associations made so far (so that repeated label letters are forced to resolve to the same types)
        :type strict: boolean
        :param strict: if ``True``, don't allow subtypes to match supertypes
        :type reversed: if ``True``, match structures contravariantly instead of covariantly (for instance, when matching arguments of functions passed as arguments)
        :rtype: bool
        :return: ``True`` if ``arg`` matches ``pat``; ``False`` otherwise
        """

        if isinstance(arg, ExceptionType):
            return False

        elif isinstance(pat, P.Null) and isinstance(arg, AvroNull):
            return True
        elif isinstance(pat, P.Boolean) and isinstance(arg, AvroBoolean):
            return True

        elif isinstance(pat, P.Int) and isinstance(arg, AvroInt):
            return True
        elif isinstance(pat, P.Long) and isinstance(arg, AvroLong):
            return True
        elif isinstance(pat, P.Float) and isinstance(arg, AvroFloat):
            return True
        elif isinstance(pat, P.Double) and isinstance(arg, AvroDouble):
            return True

        elif not strict and not reversed and \
                 isinstance(pat, P.Long) and (isinstance(arg, AvroInt) or isinstance(arg, AvroLong)):
            return True
        elif not strict and not reversed and \
                 isinstance(pat, P.Float) and (isinstance(arg, AvroInt) or isinstance(arg, AvroLong) or isinstance(arg, AvroFloat)):
            return True
        elif not strict and not reversed and \
                 isinstance(pat, P.Double) and (isinstance(arg, AvroInt) or isinstance(arg, AvroLong) or isinstance(arg, AvroFloat) or isinstance(arg, AvroDouble)):
            return True

        elif not strict and reversed and \
                 (isinstance(pat, P.Int) or isinstance(pat, P.Long)) and isinstance(arg, AvroLong):
            return True
        elif not strict and reversed and \
                 (isinstance(pat, P.Int) or isinstance(pat, P.Long) or isinstance(pat, P.Float)) and isinstance(arg, AvroFloat):
            return True
        elif not strict and reversed and \
                 (isinstance(pat, P.Int) or isinstance(pat, P.Long) or isinstance(pat, P.Float) or isinstance(pat, P.Double)) and isinstance(arg, AvroDouble):
            return True

        elif isinstance(pat, P.Bytes) and isinstance(arg, AvroBytes):
            return True
        elif isinstance(pat, P.String) and isinstance(arg, AvroString):
            return True

        elif isinstance(pat, P.Array) and isinstance(arg, AvroArray):
            return self.check(pat.items, arg.items, labelData, strict, reversed)
        elif isinstance(pat, P.Map) and isinstance(arg, AvroMap):
            return self.check(pat.values, arg.values, labelData, strict, reversed)
        elif isinstance(pat, P.Union) and isinstance(arg, AvroUnion):
            found = False
            originalLabelData = dict(labelData)
            labelData2 = dict(labelData)
            atypesPermutations = itertools.permutations(arg.types)

            atypesPermutations_isEmpty = False
            while not found and not atypesPermutations_isEmpty:
                available = OrderedDict((i, x) for i, x in enumerate(pat.types))
                try:
                    nextPermutation = next(atypesPermutations)
                except StopIteration:
                    atypesPermutations_isEmpty = True
                else:
                    anyfalse = False
                    for a in nextPermutation:
                        foundone = False
                        for i, p in list(available.items()):
                            if self.check(p, a, labelData2, True, reversed):
                                del available[i]
                                foundone = True
                                break

                        if not foundone:
                            anyfalse = True
                            break

                    if not anyfalse:
                        found = True

                    if not found:
                        labelData2 = dict(originalLabelData)

            labelData.update(labelData2)
            return found
            
        elif isinstance(pat, P.Union) and isinstance(arg, AvroType):
            return any(self.check(p, arg, labelData, strict, reversed) for p in pat.types)

        elif isinstance(pat, P.Fixed) and isinstance(arg, AvroFixed):
            if pat.fullName is not None:
                # Change in avro.schema.Name class in Py3 lead to "." at the begining of arg.fullName
                if "." == arg.fullName[0]:
                    return pat.fullName == arg.name
                else:
                    return pat.fullName == arg.fullName
            else:
                return pat.size == arg.size

        elif isinstance(pat, P.Enum) and isinstance(arg, AvroEnum):
            if pat.fullName is not None:
                if "." == arg.fullName[0]:
                    return pat.fullName == arg.name
                else:
                    return pat.fullName == arg.fullName
            else:
                return pat.symbols == arg.symbols

        elif isinstance(pat, P.Record) and isinstance(arg, AvroRecord):
            if pat.fullName is not None:
                if "." == arg.fullName[0]:
                    return pat.fullName == arg.name
                else:
                    return pat.fullName == arg.fullName
            else:
                amap = dict((x.name, x) for x in arg.fields)

                if set(pat.fields.keys()) == set(amap.keys()):
                    return all(self.check(pat.fields[k], amap[k].avroType, labelData, True, reversed) for k in list(pat.fields.keys()))
                else:
                    return False

        elif isinstance(pat, P.Fcn) and isinstance(arg, FcnType):
            return all(self.check(p, a, labelData, strict, True) for p, a in zip(pat.params, arg.params)) and \
                   self.check(pat.ret, arg.ret, labelData, strict, False)

        elif isinstance(pat, P.Wildcard):
            if pat.oneOf is None or arg in pat.oneOf:
                if not pat.label in labelData:
                    labelData[pat.label] = LabelData()
                labelData[pat.label].add(arg)
                return True
            else:
                return False

        elif isinstance(pat, P.WildRecord) and isinstance(arg, AvroRecord):
            if not pat.label in labelData:
                labelData[pat.label] = LabelData()
            labelData[pat.label].add(arg)

            amap = dict((x.name, x) for x in arg.fields)

            if set(pat.minimalFields.keys()).issubset(set(amap.keys())):
                return all(self.check(pt, amap[pn].avroType, labelData, True, reversed) for pn, pt in list(pat.minimalFields.items()))
            else:
                return False

        elif isinstance(pat, P.WildEnum):
            if not pat.label in labelData:
                labelData[pat.label] = LabelData()
            labelData[pat.label].add(arg)
            return True

        elif isinstance(pat, P.WildFixed):
            if not pat.label in labelData:
                labelData[pat.label] = LabelData()
            labelData[pat.label].add(arg)
            return True

        elif isinstance(pat, P.EnumFields) and isinstance(arg, AvroEnum):
            if not pat.label in labelData:
                labelData[pat.label] = LabelData()
            labelData[pat.label].add(arg)
            labelData[pat.wildRecord].expectedFields = arg.symbols
            return True

        else:
            return False

    def assign(self, pat, arg, assignments):
        """Apply the label assignments (e.g. "A" matched to "int", "B" matched to "string", etc.) to each parameter of the signature.

        :type pat: titus.P
        :param pat: original parameter pattern
        :type arg: titus.datatype.AvroType
        :param arg: supplied argument type
        :type assignments: dict from label letters to titus.datatype.AvroType
        :param assignments: assigned types to apply
        :rtype: titus.datatype.AvroType
        :return: resolved type for one parameter of the signature
        """

        if isinstance(pat, P.Null) and isinstance(arg, AvroNull):
            return arg
        elif isinstance(pat, P.Boolean) and isinstance(arg, AvroBoolean):
            return arg

        elif isinstance(pat, P.Int) and isinstance(arg, AvroInt):
            return AvroInt()
        elif isinstance(pat, P.Long) and (isinstance(arg, AvroInt) or isinstance(arg, AvroLong)):
            return AvroLong()
        elif isinstance(pat, P.Float) and (isinstance(arg, AvroInt) or isinstance(arg, AvroLong) or isinstance(arg, AvroFloat)):
            return AvroFloat()
        elif isinstance(pat, P.Double) and (isinstance(arg, AvroInt) or isinstance(arg, AvroLong) or isinstance(arg, AvroFloat) or isinstance(arg, AvroDouble)):
            return AvroDouble()

        elif isinstance(pat, P.Bytes) and isinstance(arg, AvroBytes):
            return arg
        elif isinstance(pat, P.String) and isinstance(arg, AvroString):
            return arg

        elif isinstance(pat, P.Array) and isinstance(arg, AvroArray):
            return AvroArray(P.mustBeAvro(self.assign(pat.items, arg.items, assignments)))
        elif isinstance(pat, P.Map) and isinstance(arg, AvroMap):
            return AvroMap(P.mustBeAvro(self.assign(pat.values, arg.values, assignments)))
        elif isinstance(pat, P.Union) and isinstance(arg, AvroUnion):
            return arg
        elif isinstance(pat, P.Union) and isinstance(arg, AvroType):
            return arg

        elif isinstance(pat, P.Fixed) and isinstance(arg, AvroFixed):
            return arg
        elif isinstance(pat, P.Enum) and isinstance(arg, AvroEnum):
            return arg
        elif isinstance(pat, P.Record) and isinstance(arg, AvroRecord):
            return arg

        elif isinstance(pat, P.Fcn) and isinstance(arg, FcnType):
            return arg

        elif isinstance(pat, P.Wildcard):
            return assignments[pat.label]
        elif isinstance(pat, P.WildRecord):
            return assignments[pat.label]
        elif isinstance(pat, P.WildEnum):
            return assignments[pat.label]
        elif isinstance(pat, P.WildFixed):
            return assignments[pat.label]
        elif isinstance(pat, P.EnumFields):
            return assignments[pat.label]

        else:
            raise Exception

    def assignRet(self, pat, assignments):
        """Apply the label assignments (e.g. "A" matched to "int", "B" matched to "string" etc.) to the return pattern.

        :type pat: titus.P
        :param pat: original return pattern
        :type assignments: dict from label letters to titus.datatype.AvroType
        :param assignments: assigned types to apply
        :rtype: titus.datatype.AvroType
        :return: resolved type for the return value of the signature
        """

        if isinstance(pat, P.Null):
            return AvroNull()
        elif isinstance(pat, P.Boolean):
            return AvroBoolean()
        elif isinstance(pat, P.Int):
            return AvroInt()
        elif isinstance(pat, P.Long):
            return AvroLong()
        elif isinstance(pat, P.Float):
            return AvroFloat()
        elif isinstance(pat, P.Double):
            return AvroDouble()
        elif isinstance(pat, P.Bytes):
            return AvroBytes()
        elif isinstance(pat, P.String):
            return AvroString()

        elif isinstance(pat, P.Array):
            return AvroArray(self.assignRet(pat.items, assignments))
        elif isinstance(pat, P.Map):
            return AvroMap(self.assignRet(pat.values, assignments))
        elif isinstance(pat, P.Union):
            return AvroUnion([self.assignRet(x, assignments) for x in pat.types])

        elif isinstance(pat, P.Fixed):
            return P.toType(pat)
        elif isinstance(pat, P.Enum):
            return P.toType(pat)
        elif isinstance(pat, P.Record):
            return P.toType(pat)
        elif isinstance(pat, P.Fcn):
            return P.toType(pat)

        elif isinstance(pat, P.Wildcard):
            return assignments[pat.label]
        elif isinstance(pat, P.WildRecord):
            return assignments[pat.label]
        elif isinstance(pat, P.WildEnum):
            return assignments[pat.label]
        elif isinstance(pat, P.WildFixed):
            return assignments[pat.label]
        elif isinstance(pat, P.EnumFields):
            return assignments[pat.label]

        else:
            raise Exception(repr(pat))

def toText(p, alreadyLabeled=None):
    """Render a pattern as human-readable text.

    :type p: titus.P
    :param p: the pattern
    :type alreadyLabeled: set of strings or ``None``
    :param alreadyLabeled: used to avoid recursion in recursive types
    :rtype: string
    :return: text representation of the pattern
    """

    if alreadyLabeled is None:
        alreadyLabeled = set()
    if isinstance(p, P.Null):
        return "null"
    elif isinstance(p, P.Boolean):
        return "boolean"
    elif isinstance(p, P.Int):
        return "int"
    elif isinstance(p, P.Long):
        return "long"
    elif isinstance(p, P.Float):
        return "float"
    elif isinstance(p, P.Double):
        return "double"
    elif isinstance(p, P.Bytes):
        return "bytes"
    elif isinstance(p, P.Fixed):
        if p.fullName is None:
            return "fixed(size: {0})".format(p.size)
        else:
            return "fixed(size: {0}, name: {1})".format(p.size, p.fullName)
    elif isinstance(p, P.String):
        return "string"
    elif isinstance(p, P.Enum):
        if p.fullName is None:
            return "fixed(symbols: {0})".format(", ".join(p.symbols))
        else:
            return "fixed(symbols: {0}, name: {1})".format(", ".join(p.symbols), p.fullName)
    elif isinstance(p, P.Array):
        return "array of " + toText(p.items, alreadyLabeled)
    elif isinstance(p, P.Map):
        return "map of " + toText(p.values, alreadyLabeled)
    elif isinstance(p, P.Record):
        if p.fullName is None:
            return "record (fields: {" + ", ".join(n + ": " + toText(f, alreadyLabeled) for n, f in list(p.fields.items())) + "})"
        else:
            return "record (fields: {" + ", ".join(n + ": " + toText(f, alreadyLabeled) for n, f in list(p.fields.items())) + ", name: " + p.fullName + "})"
    elif isinstance(p, P.Union):
        return "union of {" + ", ".join(toText(x, alreadyLabeled) for x in p.types) + "}"
    elif isinstance(p, P.Fcn):
        return "function of (" + ", ".join(toText(x, alreadyLabeled) for x in p.params) + ") -> " + toText(p.ret, alreadyLabeled)
    elif isinstance(p, P.Wildcard) and p.label in alreadyLabeled:
        return p.label
    elif isinstance(p, P.WildEnum) and p.label in alreadyLabeled:
        return p.label
    elif isinstance(p, P.WildFixed) and p.label in alreadyLabeled:
        return p.label
    elif isinstance(p, P.WildRecord) and p.label in alreadyLabeled:
        return p.label
    elif isinstance(p, P.EnumFields) and p.label in alreadyLabeled:
        return p.label
    elif isinstance(p, P.Wildcard):
        alreadyLabeled.add(p.label)
        if len(p.oneOf) == 0:
            return "any " + p.label
        else:
            return "any " + p.label + " of {" + ", ".join(toTextType(x) for x in p.oneOf) + "}"
    elif isinstance(p, P.WildEnum):
        alreadyLabeled.add(p.label)
        return "any enum " + p.label
    elif isinstance(p, P.WildFixed):
        alreadyLabeled.add(p.label)
        return "any fixed " + p.label
    elif isinstance(p, P.WildRecord):
        alreadyLabeled.add(p.label)
        return "any record " + p.label + ("" if len(p.minimalFields) == 0 else " with fields {" + ", ".join(n + ": " + toText(f, alreadyLabeled) for n, f in list(p.minimalFields.items())) + "}")
    elif isinstance(p, P.EnumFields):
        alreadyLabeled.add(p.label)
        return "enum " + p.label + " of fields of " + p.wildRecord

def toTextType(x):
    """Render a type as human-readable text.

    Note that this function is incomplete, but it covers all the cases seen in any PFA functions.

    :type x: titus.datatype.AvroType
    :param x: the data type
    :rtype: string
    :return: text representation of the data type
    """

    if isinstance(x, AvroNull):
        return "null"
    elif isinstance(x, AvroBoolean):
        return "boolean"
    elif isinstance(x, AvroInt):
        return "int"
    elif isinstance(x, AvroLong):
        return "long"
    elif isinstance(x, AvroFloat):
        return "float"
    elif isinstance(x, AvroDouble):
        return "double"
    elif isinstance(x, AvroBytes):
        return "bytes"
    elif isinstance(x, AvroString):
        return "string"
    elif isinstance(x, AvroArray):
        return "array of " + toTextType(x.items)
    elif isinstance(x, AvroMap):
        return "map of " + toTextType(x.values)



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import os
import os.path
import readline
import traceback
import re
import sys
import subprocess
import glob
import math

import titus.inspector.parser as parser
from titus.reader import jsonToAst
from titus.genpy import PFAEngine
from titus.errors import AvroException, SchemaParseException, PFAException

CONFIG_DIRECTORY = "~/.pfa"
CONFIG_DIRECTORY_EXISTS = True

if not os.path.exists(os.path.expanduser(CONFIG_DIRECTORY)):
    if input("Create {0} for configuration files? (Y/n): ".format(CONFIG_DIRECTORY)).upper().strip() in ("Y", ""):
        os.mkdir(os.path.expanduser(CONFIG_DIRECTORY))
    else:
        CONFIG_DIRECTORY_EXISTS = False

CONFIG_COMPLETER_DELIMS = " \t[],:="
readline.set_completer_delims(CONFIG_COMPLETER_DELIMS)

readline.parse_and_bind("tab: complete")
readline.parse_and_bind(r'"\eh": backward-kill-word')
readline.parse_and_bind(r'"\ep": history-search-backward')
readline.parse_and_bind(r'"\en": history-search-forward')

class TabCompleter(object):
    """Handles tab-completion in pfainspector."""

    def __init__(self, mode):
        """:type mode: titus.inspector.defs.Mode
        :param mode: the pfainspector mode in which this tab completer is active
        """

        self.mode = mode
        self.candidates = []

    def complete(self, text, state):
        """:type text: string
        :param text: partial text to complete
        :type state: integer
        :param state: the number of times the user has pressed tab. If ``0``, generate a new list of candidates; otherwise, use the old one
        :rtype: list of strings
        :return: set of completions
        """

        if state == 0:
            line = readline.get_line_buffer()
            begin = readline.get_begidx()
            end = readline.get_endidx()
            established = line[:begin]
            active = line[begin:end]
            self.candidates = self.mode.complete(established, active)
        try:
            return self.candidates[state]
        except IndexError:
            return None

class InspectorError(Exception):
    """Exception encountered in pfainspector."""
    pass

class Gadget(object):
    """Trait for pfainspector gadgets."""
    def __init__(self, mode):
        self.commandGroup = None

def do(*args):
    """Helper function for chaining function calls in a Python lambda expression.

    Part of a poor-man's functional programming suite used to define several pfainspector commands as one-liners.

    :type args: anything
    :param args: functions that have already been evaluated
    :rtype: anything
    :return: the last argument
    """
    return args[-1]

def maybe(action, exceptions):
    """Helper function for try-catch logic in a Python lambda expression.

    Part of a poor-man's functional programming suite used to define several pfainspector commands as one-liners.

    :type action: callable
    :param action: function to call
    :type exceptions: list of (class, string)
    :param exceptions: exception classes to catch and their corresponding pfainspector messages
    :rtype: anything
    :return: result of calling ``action``
    """

    try:
        action()
    except Exception as err:
        for cls, message in exceptions:
            if isinstance(err, cls):
                raise InspectorError(message)
        raise

def switch(*pairs):
    """Helper function for cond-logic in a Python lambda expression.

    Part of a poor-man's functional programming suite used to define several pfainspector commands as one-liners.

    :type pairs: callables
    :param pairs: sequence of predicate1, consequent1, predicate2, consequent2, ..., alternate; the predicates will each be called in turn until one returns ``True``, and then the corresponding consequent will be called *or* the alternate will be called if none of the predicates return ``True``
    """

    if len(pairs) % 2 != 1 or len(pairs) < 3:
        raise TypeError
    for predicate, consequent in zip(pairs[:-1][::2], pairs[:-1][1::2]):
        if callable(predicate):
            predicate = predicate()
        if predicate:
            if callable(consequent):
                consequent = consequent()
            return consequent
    alterante = pairs[-1]
    if callable(alterante):
        alterante = alterante()
    return alterante

def exception(x):
    """Helper function for raising an exception in a Python lambda expression.

    Part of a poor-man's functional programming suite used to define several pfainspector commands as one-liners.

    :type x: ``Exception``
    :param x: exception to raise (from a function call)
    """

    raise x

def getwords(text):
    """Parse a (partial?) command line into a syntax tree.

    :type text: string
    :param text: text to parse
    :rtype: list of titus.inspector.parser.Ast
    :return: abstract syntax tree for a pfainspector command line
    """

    try:
        return parser.parser.parse(text)
    except parser.ParserError as err:
        raise InspectorError(str(err))

def getcomplete(established):
    """Get the result of tab completion.

    :type established: string
    :param established: text that has been established and is not subject to completion (though it may influence completion as a context)
    :rtype: list of titus.inspector.parser.Ast
    :return: abstract syntax tree of the established part of a pfainspector command line
    """

    if established.strip() == "":
        return []
    else:
        return parser.parser.parse(established)  # let it fail; completer ignores exceptions

def pathcomplete(established, active):
    """Tab completion routine for filesystem paths.

    :type established: string
    :param established: text that has been established and is not subject to completion (everything up to the last ``/``)
    :type active: string
    :param active: text to be completed (anything after the last ``/``)
    :rtype: list of strings
    :return: list of possible completions
    """

    base, last = os.path.split(os.path.expanduser(active))
    if base == "":
        base = "."
    if os.path.isdir(base):
        if last == "":
            condition = lambda x: not x.startswith(".")
        else:
            condition = lambda x: x.startswith(last)
        def finish(x):
            if " " in x:
                quote = '"'
            else:
                quote = ""
            if os.path.isdir(x):
                return quote + os.path.join(x, "")
            elif os.path.exists(x):
                return quote + x + " "
        return [finish(os.path.join(base, x) if base != "." else x) for x in os.listdir(base) if condition(x)]

def extcomplete(node, items):
    """Tab completion routine for JSON extraction (everything after an opening square bracket).

    :type established: string
    :param established: text that has been established and is not subject to completion (everything up to the last ``,``)
    :type active: string
    :param active: text to be completed (anything after the last ``,``)
    :rtype: list of strings
    :return: list of possible completions
    """

    for item in items:
        if isinstance(item, (parser.Word, parser.String)) and isinstance(node, dict):
            if item.text in node:
                node = node[item.text]
            else:
                return []
        elif isinstance(item, parser.Integer) and isinstance(node, (list, tuple)):
            if item.num < len(node):
                node = node[item.num]
            else:
                return []
        else:
            return []

    def quote(x):
        if " " in x:
            return '"' + x + '"'
        else:
            return x

    if isinstance(node, dict):
        return [quote(x) + (", " if isinstance(node[x], (list, tuple, dict)) else "]") for x in node]

    elif isinstance(node, (list, tuple)):
        formatter = "%%0%dd" % int(math.ceil(math.log10(len(node))))
        return [quote(formatter % x) + (", " if isinstance(node[x], (list, tuple, dict)) else "]") for x in range(len(node))]

    else:
        return []

def extaction(args0, node, items):
    """Action for pfainspector extensions (depend on specific commands).

    :type args0: titus.inspector.parser.Extract
    :param args0: abstract syntax tree representation of the object to extract (word before opening square bracket)
    :type node: Pythonized JSON
    :param node: JSON node from which to extract subobjects
    :type items: (integer, titus.inspector.parser.FilePath)
    :param items: extraction path (everything between square brakets, possibly still open)
    :rtype: Pythonized JSON
    :return: result of extracting subobjects
    """

    for index, item in enumerate(items):
        if isinstance(item, (parser.Word, parser.String)):
            try:
                node = node[item.text]
            except (KeyError, TypeError):
                raise InspectorError("{0} has no key {1}".format(args0.strto(index), str(item)))
        elif isinstance(item, parser.Integer):
            try:
                node = node[item.num]
            except (IndexError, TypeError):
                raise InspectorError("{0} has no index {1}".format(args0.strto(index), str(item)))
        else:
            raise InspectorError("syntax error: {0} should not appear in extractor".format(item))

    return node

def run(command, *args):
    """Helper function to run a subprocess in a Python lambda expression.

    :type command: string
    :param command: external command to run
    :type args: strings
    :param args: arguments for the external command
    :rtype: string
    :return: result of running the command
    """

    whole = [command]
    for arg in args:
        g = glob.glob(arg)
        if len(g) > 0:
            whole.extend(g)
        else:
            whole.append(arg)
    proc = subprocess.Popen(whole, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    returnCode = proc.wait()
    output = proc.stdout.read()
    if returnCode != 0:
        output += "\nsubprocesses failed with exit code {0}".format(returnCode)
    return output

def pipe(command):
    """Helper function to create a subprocess.Popen as a pipe.

    :type command: list of strings
    :param command: external command with arguments
    :rtype: subprocess.Popen object
    :return: piped subprocess (for further processing)
    """
    return subprocess.Popen(command, stdin=subprocess.PIPE)

def pipewait(proc):
    """Helper function to wait for a subprocess.Popen to complete.

    :type proc: subprocess.Popen
    :param proc: process to wait for
    :rtype: ``None``
    :return: nothing; if the process's return code is not ``0``, a warning message is printed to standard output
    """

    proc.stdin.close()
    returnCode = proc.wait()
    if returnCode != 0:
        print("\nsubprocesses failed with exit code {0}".format(returnCode))

class Model(object):
    """A loaded JSON or PFA file.

    Always has an ``obj`` member, representing the Pythonized JSON that was loaded from the file.

    Lazy-evaluates an ``engineConfig`` member the first time it is requested. This is a titus.pfaast.EngineConfig representing the abstract syntax tree of the PFA file. If the PFA file contains an error, attempting to access the ``engineConfig`` property yields an exception (titus.errors.AvroException, titus.errors.SchemaParseException, or titus.errors.PFAException).

    Lazy-evaluates an ``engine`` member the first time it is requested. This is a titus.genpy.PFAEngine representing an executable scoring engine. If the PFA file contains an error, attempting to access the ``engineConfig`` property yields an exception (titus.errors.AvroException, titus.errors.SchemaParseException, or titus.errors.PFAException).
    """

    def __init__(self, obj):
        self.obj = obj
        self._engineConfig = None
        self._engine = None

    def reset(self):
        self._engineConfig = None
        self._engine = None
        
    @property
    def engineConfig(self):
        if self._engineConfig is None:
            try:
                self._engineConfig = jsonToAst(self.obj)
            except (AvroException, SchemaParseException, PFAException) as err:
                raise InspectorError(str(err))
        return self._engineConfig

    @property
    def engine(self):
        if self._engine is None:
            try:
                self._engine, = PFAEngine.fromAst(self.engineConfig)
            except (AvroException, SchemaParseException, PFAException) as err:
                raise InspectorError(str(err))
        return self._engine
        
    def __repr__(self):
        return "Model(" + repr(self.obj) + ")"
    
class Mode(object):
    """A mode of operation for the pfainspector.

    This concrete base class is a read-eval-print loop for responding to the user's commands.
    """

    def __init__(self):
        """Create the main mode.

        If titus.inspector.defs.CONFIG_DIRECTORY_EXISTS is ``True``, get the readline history file from the user's titus.inspector.defs.CONFIG_DIRECTORY.
        """

        if CONFIG_DIRECTORY_EXISTS:
            self.historyPath = os.path.join(os.path.expanduser(CONFIG_DIRECTORY), self.historyFileName)
            if not os.path.exists(self.historyPath):
                open(self.historyPath, "w").close()

            self.active = True
            self.tabCompleter = TabCompleter(self)
            readline.read_history_file(self.historyPath)
            readline.set_completer(self.tabCompleter.complete)

            def writehistory():
                if self.active:
                    readline.write_history_file(self.historyPath)
            atexit.register(writehistory)

    def loop(self):
        """Main loop: attempts to evaluate commands.

        An ``EOFError`` exception (user typed control-D on an empty line) quits the pfainspector.

        A ``KeyboardInterrupt`` exception (user typed control-C at any time) jumps to an empty command-line state.

        Any titus.inspector.defs.InspectorError results in the error's message being printed on the screen.

        Any other exception results in a full stack trace being printed on the screen.
        """

        while True:
            try:
                line = input(self.prompt)
                if line.strip() != "":
                    self.action(line)
            except EOFError:
                print()
                sys.exit(0)
            except KeyboardInterrupt:
                print("(use control-D or exit to quit)")
            except InspectorError as err:
                print(err)
            except Exception as err:
                traceback.print_exc()

    def pause(self):
        """Write the current history to the history file and stop readline."""
        readline.write_history_file(self.historyPath)
        readline.clear_history()
        readline.set_completer()
        self.active = False

    def resume(self):
        """Read the history file and restart readline."""
        self.active = True
        readline.read_history_file(self.historyPath)
        readline.set_completer(self.tabCompleter.complete)

class Command(object):
    """Trait for pfainspector commands."""

    def __init__(self):
        pass

    def syntaxError(self):
        """General behavior for syntax errors in the command parser: raise a titus.inspector.defs.InspectorError."""

        if self.syntax is None:
            raise InspectorError("syntax error in {0}".format(self.name))
        else:
            raise InspectorError("syntax error, should be: {0}".format(self.syntax))

class SimpleCommand(Command):
    """A pfainspector command whose action can be expressed as a Python lambda expression."""

    def __init__(self, name, action, minargs=None, maxargs=None, completer=None, help=None, syntax=None):
        """Create a SimpleCommand.

        :type name: string
        :param name: name of the command
        :type action: callable
        :param action: action to perform, usually supplied as a Python lambda expression.
        :type minargs: non-negative integer or ``None``
        :param minargs: if provided, a minimum legal number of arguments for the command
        :type maxargs: non-negative integer or ``None``
        :param maxargs: if provided, a maximum legal number of arguments for the command
        :type completer: callable
        :param completer: function to call to complete the arguments of the command
        :type help: string
        :param help: message to the user describing the purpose of this command
        :type syntax: string
        :param syntax: message to the user specifying the syntax of this command
        """

        self.name = name
        self._action = action
        self.minargs = minargs
        self.maxargs = maxargs
        self.completer = completer
        self.syntax = syntax
        self.help = help + "\n    " + syntax

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        out = []
        if self.completer is None:
            pass
        elif callable(self.completer):
            out.extend(self.completer(established, active))
        else:
            for completer in self.completers:
                out.extend(completer(established, active))
        return out

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if self.help is not None and len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            for arg in args:
                if not isinstance(arg, parser.FilePath):
                    raise InspectorError("arguments must all be words, not {0}".format(arg))

            if self.minargs is not None and len(args) < self.minargs:
                raise InspectorError("too few arguments (at least {0} are required)".format(self.minargs))
            if self.maxargs is not None and len(args) < self.maxargs:
                raise InspectorError("too many arguments (at most {0} are required)".format(self.maxargs))

            out = self._action(*[x.text for x in args])
            if out is not None:
                print(out)

class CommandGroup(Command):
    """A pfainspector command that defers to a group of subcommands."""

    def __init__(self, name, commands):
        """:type name: string
        :param name: name of the group
        :type commands: list of titus.inspector.defs.Command
        :param commands: commands in this group
        """

        self.name = name
        self.commands = dict((x.name, x) for x in commands)
        if name is None:
            self.help = "Commands:\n"
        else:
            self.help = "{0} gadget (type '{1} help' for details)\nSubcommands under {2}:\n".format(name, name, name)
        for x in commands:
            self.help += "    {0:<20s} {1}\n".format(x.name, x.help.split("\n")[0] if x.help is not None else "")
        self.help = self.help.strip()

    def complete(self, established, active):
        """Handle tab-complete for the command group: either expanding the subcommand name or deferring to the subcommand's ``complete`` method.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        word = re.match(r'''\s*([A-Za-z][A-Za-z0-9_]*)\s*''', established)
        if word is None:
            return [x + " " for x in self.commands if x.startswith(active)]
        else:
            try:
                command = self.commands[word.groups()[0]]
            except KeyError:
                return []
            else:
                return command.complete(established[word.end():], active)

    def action(self, args):
        """Perform the action associated associated with this command group: descend into the subcommand and call its ``action`` method.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 0:
            if self.name is not None:
                raise InspectorError("command {0} requires a subcommand".format(self.name))

        elif len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)

        elif isinstance(args[0], parser.Word):
            try:
                command = self.commands[args[0].text]
            except KeyError:
                if args[0] == parser.Word("help"):
                    raise InspectorError("help commands should end with the word 'help'")
                if self.name is None:
                    raise InspectorError("no command named {0}".format(args[0].text))
                else:
                    raise InspectorError("command {0} has no subcommand {1}".format(self.name, args[0].text))
            else:
                command.action(args[1:])

        else:
            if self.name is None:
                raise InspectorError("command must be a word, not {0}".format(args[0]))
            else:
                raise InspectorError("subcommand of {0} must be a word, not {1}".format(self.name, args[0]))



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn
prefix = "m.special."

class NChooseK(LibFcn):
    name = prefix + "nChooseK"
    sig = Sig([{"n": P.Int()}, {"k": P.Int()}], P.Int())
    errcodeBase = 36000
    def __call__(self, state, scope, pos, paramTypes, n, k):
        try:
            if n == k: raise Exception
            return int(nChooseK(n, k))
        except:
            raise PFARuntimeException("domain error", self.errcodeBase + 0, self.name, pos)
provide(NChooseK())

class LnBeta(LibFcn):
    name = prefix + "lnBeta"
    sig = Sig([{"a": P.Double()}, {"b": P.Double()}], P.Double())
    errcodeBase = 36010
    def __call__(self, state, scope, pos, paramTypes, a, b):
        try:
            return logBetaFunction(a,b)
        except:
            raise PFARuntimeException("domain error", self.errcodeBase + 0, self.name, pos)
provide(LnBeta())

class Erf(LibFcn):
    name = prefix + "erf"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 36020
    def __call__(self, state, scope, pos, paramTypes, a):
        try:
            return math.erf(a)
        except:
            raise PFARuntimeException("domain error", self.errcodeBase + 0, self.name, pos)
provide(Erf())

class Erfc(LibFcn):
    name = prefix + "erfc"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 36040
    def __call__(self, state, scope, pos, paramTypes, a):
        try:
            return math.erfc(a)
        except:
            raise PFARuntimeException("domain error", self.errcodeBase + 0, self.name, pos)
provide(Erfc())

class LnGamma(LibFcn):
    name = prefix + "lnGamma"
    sig = Sig([{"x": P.Double()}], P.Double())
    errcodeBase = 36050
    def __call__(self, state, scope, pos, paramTypes, a):
        if math.isinf(a) and a > 0:
            return float("nan")
        try:
            if a <= 0: raise Exception
            else:
                return math.lgamma(a)
        except:
            raise PFARuntimeException("domain error", self.errcodeBase + 0, self.name, pos)
provide(LnGamma())

#class IncompleteBeta(LibFcn):
#    name = prefix + "incompletebetafcn"
#    sig = Sig([{"x": P.Double()}, {"a": P.Double()}, {"b": P.Double()}], P.Double())
#    errcodeBase = 36060
#    def __call__(self, state, scope, pos, paramTypes, x, a, b):
#        return incompleteBetaFunction(x, a, b)
#provide(IncompleteBeta())

#class IncompleteBetaInv(LibFcn):
#    name = prefix + "incompletebetafcninv"
#    sig = Sig([{"p": P.Double()}, {"a": P.Double()}, {"b": P.Double()}], P.Double())
#    errcodeBase = 36070
#    def __call__(self, state, scope, pos, paramTypes, p, a, b):
#        return inverseIncompleteBetaFunction(p, a, b)
#provide(IncompleteBetaInv())

# http://mathworld.wolfram.com/RegularizedGammaFunction.html
#class RegularizedGammaP(LibFcn):
#    name = prefix + "regularizedgammapfcn"
#    sig = Sig([{"x": P.Double()}, {"a": P.Double()}], P.Double())
#    errcodeBase = 36080
#    def __call__(self, state, scope, pos, paramTypes, x, a):
#        return regularizedGammaP(a, x)
#provide(RegularizedGammaP())

# http://mathworld.wolfram.com/RegularizedGammaFunction.html
#class RegularizedGammaQ(LibFcn):
#    name = prefix + "regularizedgammaqfcn"
#    sig = Sig([{"x": P.Double()}, {"a": P.Double()}], P.Double())
#    errcodeBase = 36090
#    def __call__(self, state, scope, pos, paramTypes, x, a):
#        return regularizedGammaQ(a, x)
#provide(RegularizedGammaQ())

#########################################################################################
### The actual functions ################################################################
#########################################################################################

def logBetaFunction(a, b):
    if (a <= 0.0) or (b <= 0.0):
        raise Exception
    else:
        return math.lgamma(a) + math.lgamma(b) - math.lgamma(a + b)

# numerical recipes in C
def incompleteBetaFunction(x,a,b):
    try:
        lbeta = math.lgamma(a + b) - math.lgamma(a) - math.lgamma(b) \
                + a * math.log(x) + b * math.log(1.0 - x)
    except ValueError:
        lbeta = float("nan")
    if (x < (a + 1)/(a + b + 2)):
        return math.exp(lbeta) * contFractionBeta(a,b,x)/a
    else:
        return 1 - math.exp(lbeta) * contFractionBeta(b,a,1.-x)/b

# numerical recipes in C
def inverseIncompleteBetaFunction(p, a, b):
    a1 = a - 1.0
    b1 = b - 1.0
    ERROR = 1.0e-10

    if (p <= 0.0):
        return 0.0
    elif (p >= 1.0):
        return 1.0;
    elif (a >= 1.0) and (b >= 1.0):
        if (p < 0.5):
            pp = p
        else:
            pp = 1.0 - p;
        t = math.sqrt(-2.0*math.log(pp))
        x = (2.30753+t*0.27061)/(1.0+t*(0.99229+t*0.04481)) - t
        if (p < 0.5):
            x = -x
        al = ((x*x)-3.0)/6.0;
        h = 2.0/(1.0/(2.0*a-1.0)+1.0/(2.0*b-1.0));
        w = (x*math.sqrt(al+h)/h)-(1.0/(2.0*b-1)-1.0/(2.0*a-1.0))*(al+5.0/6.0-2.0/(3.0*h))
        x = a/(a+b*math.exp(2.0*w))
    else:
        lna = math.log(a/(a+b))
        lnb = math.log(b/(a+b))
        t = math.exp(a*lna)/a
        u = math.exp(b*lnb)/b
        w = t + u
        if (p < t/w):
            x = math.pow(a*w*p,1.0/a)
        else:
            x = 1.0 - math.pow(b*w*(1.0-p), 1.0/b)
    afac = -math.lgamma(a)-math.lgamma(b)+math.lgamma(a+b)
    j = 0
    for i in range(10):
        if (x == 0.0) or (x == 1.0):
            return x;
        err = incompleteBetaFunction(x,a,b) - p

        t = math.exp(a1*math.log(x)+b1*math.log(1.0-x) + afac)
        u = err/t
        t = u/(1.0-0.5*min(1.0,u*(a1/x - b1/(1.0-x))))
        x -= t
        if (x <= 0.0):
            x = 0.5*(x + t)
        if (x >= 1.0):
            x = 0.5*(x + t + 1.0)
        if (math.fabs(t) < ERROR*x) and (j > 0):
            break
    return x

# numerical recipes in C
# used by incompleteBeta and incompleteBetaComp
def contFractionBeta(a,b,x, maxIter = 1000, epsilon=1e-15):
    bm = 1.0
    az = 1.0
    am = 1.0
    qab = a + b
    qap = a + 1.0
    qam = a - 1.0
    bz = 1.0 - qab*x/qap
    for i in range(0,maxIter):
        em = i + 1
        tem = em + em
        d = em*(b-em)*x/((qam+tem)*(a+tem))
        ap = az + d*am
        bp = bz+d*bm
        d = -(a+em)*(qab+em)*x/((qap+tem)*(a+tem))
        app = ap+d*az
        bpp = bp+d*bz
        aold = az
        am = ap/bpp
        bm = bp/bpp
        az = app/bpp
        bz = 1.0
        if (abs(az-aold) < (epsilon*abs(az))):
            return az
    raise Exception

# http://grepcode.com/file/repo1.maven.org/maven2/org.apache.commons/commons-math3/3.0/org/apache/commons/math3/special/Gamma.java#Gamma.regularizedGammaP%28double%2Cdouble%29
def regularizedGammaP(a, x):
    if (x < 0.0) or (a <= 0.0):
        raise Exception
    maxIter = 1000
    epsilon = 1e-15
    n  = 0.0
    an = 1.0/a
    total = an
    while (abs(an/total) > epsilon) and \
          (n < maxIter) and \
          (total < float("inf")):
        n = n + 1.0
        an = an * (x / (a + n))
        total = total + an
    if n >= maxIter:
        raise Exception
    elif total == float("inf"):
        ans = 1.0
    else:
        ans = math.exp(-x + (a*math.log(x)) - math.lgamma(a)) * total
    return ans

# http://www.johnkerl.org/python/sp_funcs_m.py.txt
def regularizedGammaQ(a, x):
    """Incomplete gamma function."""
    if (x < 0.0 or a <= 0.0):
        raise Exception
    if (x < a+1.0):
        return 1.0 - gser(a,x)[0]
    else:
        return gcf(a,x)[0]

# http://www.johnkerl.org/python/sp_funcs_m.py.txt
# used by incompleteGamma
def gser(a, x, itmax=700, eps=3.0e-7):
    """Series approximation to the incomplete gamma function."""
    gln = math.lgamma(a)
    if (x < 0.0):
        raise Exception
    if (x == 0.0):
        return [0.0]
    ap = a
    total = 1.0 / a
    delta = total
    n = 1
    while n <= itmax:
        ap = ap + 1.0
        delta = delta * x / ap
        total = total + delta
        if (abs(delta) < abs(total)*eps):
            return [total * math.exp(-x + a*math.log(x) - gln), gln]
        n = n + 1
    raise Exception

# http://www.johnkerl.org/python/sp_funcs_m.py.txt
# used by incompleteGamma
def gcf(a, x, itmax=200, eps=3.0e-7):
    """Continued fraction approximation of the incomplete gamma function."""
    gln = math.lgamma(a)
    gold = 0.0
    a0 = 1.0
    a1 = x
    b0 = 0.0
    b1 = 1.0
    fac = 1.0
    n = 1
    while n <= itmax:
        an = n
        ana = an - a
        a0 = (a1 + a0*ana)*fac
        b0 = (b1 + b0*ana)*fac
        anf = an*fac
        a1 = x*a0 + anf*a1
        b1 = x*b0 + anf*b1
        if (a1 != 0.0):
            fac = 1.0 / a1
            g = b1*fac
            if (abs((g-gold)/g) < eps):
                return (g*math.exp(-x+a*math.log(x)-gln), gln)
            gold = g
        n = n + 1
    raise Exception

def nChooseK(n, k):
    # is n an integer?
    nInt = (math.floor(n) == n)
    if n == k or k == 0:
        return 1
    if (n < k) or (k < 0):
        raise Exception
    if (nInt) and (n < 0.0):
        b = pow(-1.0, k) * math.exp(math.lgamma(abs(n + k)) \
                                  - math.lgamma(k + 1.0)    \
                                  - math.lgamma(abs(n)))
        return round(b)
    if (n >= k):
        b = math.exp(math.lgamma(n + 1.0) - math.lgamma(k + 1.0) \
                   - math.lgamma(n - k + 1.0))
        return round(b)
    if not (nInt) and (n < k):
        b = (1.0/math.pi) * math.exp(math.lgamma(n + 1.0) \
                                   - math.lgamma(k + 1)   \
                                   + math.lgamma(k - n)   \
                   + math.log(math.sin(math.pi * (n - k + 1.0))))
        return round(b)
    return 0.0



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "model.tree."

#################################################################### 

def simpleComparison(paramTypes, datum, comparison, missingOperators, parser, code1, code2, fcnName, pos):
    field = comparison["field"]
    fieldValue = datum[field]
    operator = comparison["operator"]
    value = comparison["value"]

    fieldValueType = [x for x in paramTypes[0]["fields"] if x["name"] == field][0]["type"]
    valueType = [x for x in paramTypes[1]["fields"] if x["name"] == "value"][0]["type"]

    if not missingOperators:
        if isinstance(fieldValueType, (list, tuple)):
            withoutNull = [x for x in fieldValueType if x != "null" and x != {"type": "null"}]
            if len(withoutNull) == 1:
                fieldValueType = withoutNull[0]
            else:
                fieldValueType = withoutNull

            if isinstance(fieldValue, dict) and len(fieldValue) == 1 and any(parser.getAvroType(x).name == list(fieldValue.keys())[0] for x in withoutNull):
                fieldValue, = list(fieldValue.values())
        
    fieldValueType = parser.getAvroType(fieldValueType)
    valueType = parser.getAvroType(valueType)

    if operator == "alwaysTrue":
        return True
    elif operator == "alwaysFalse":
        return False
    elif missingOperators and operator == "isMissing":
        return (fieldValue is None)
    elif missingOperators and operator == "notMissing":
        return (fieldValue is not None)

    elif operator == "in" or operator == "notIn":
        if isinstance(valueType, AvroArray) and valueType.items.accepts(fieldValueType):
            pass
        elif isinstance(valueType, AvroUnion) and any(isinstance(x, AvroArray) and x.items.accepts(fieldValueType) for x in valueType.types):
            pass
        else:
            raise PFARuntimeException("bad value type", code1, fcnName, pos)

        if isinstance(value, dict) and len(value) == 1 and list(value.keys()) == ["array"]:
            value, = list(value.values())

        if not missingOperators:
            if fieldValue is None:
                return None

        containedInList = (fieldValue in value)
        if operator == "in":
            return containedInList
        else:
            return not containedInList

    else:
        if not valueType.accepts(fieldValueType):
            if isinstance(valueType, (AvroInt, AvroLong, AvroFloat, AvroDouble)) and \
               isinstance(fieldValueType, (AvroInt, AvroLong, AvroFloat, AvroDouble)):
                pass
            else:
                raise PFARuntimeException("bad value type", code1, fcnName, pos)

        if not missingOperators and fieldValue is None:
            return None

        if isinstance(fieldValueType, (AvroInt, AvroLong, AvroFloat, AvroDouble)):
            if isinstance(value, dict) and (list(value.keys()) == ["int"] or list(value.keys()) == ["long"] or list(value.keys()) == ["float"] or list(value.keys()) == ["double"]):
                value, = list(value.values())
            if not isinstance(value, (int, float)):
                raise PFARuntimeException("bad value type", code1, fcnName, pos)

            if operator == "<=":
                return fieldValue <= value
            elif operator == "<":
                return fieldValue < value
            elif operator == ">=":
                return fieldValue >= value
            elif operator == ">":
                return fieldValue > value
            elif operator == "==":
                return fieldValue == value
            elif operator == "!=":
                return fieldValue != value
            else:
                raise PFARuntimeException("invalid comparison operator", code2, fcnName, pos)

        else:
            if operator == "<=":
                return compare(valueType, fieldValue, value) <= 0
            elif operator == "<":
                return compare(valueType, fieldValue, value) < 0
            elif operator == ">=":
                return compare(valueType, fieldValue, value) >= 0
            elif operator == ">":
                return compare(valueType, fieldValue, value) > 0
            elif operator == "==":
                return compare(valueType, fieldValue, value) == 0
            elif operator == "!=":
                return compare(valueType, fieldValue, value) != 0
            else:
                raise PFARuntimeException("invalid comparison operator", code2, fcnName, pos)

class SimpleTest(LibFcn):
    name = prefix + "simpleTest"
    sig = Sig([{"datum": P.WildRecord("D", {})}, {"comparison": P.WildRecord("T", {"field": P.EnumFields("F", "D"), "operator": P.String(), "value": P.Wildcard("V")})}], P.Boolean())
    errcodeBase = 32000
    def __call__(self, state, scope, pos, paramTypes, datum, comparison):
        return simpleComparison(paramTypes, datum, comparison, True, state.parser, self.errcodeBase + 1, self.errcodeBase + 0, self.name, pos)
provide(SimpleTest())

class CompoundTest(LibFcn):
    name = prefix + "compoundTest"
    sig = Sig([{"datum": P.WildRecord("D", {})}, {"operator": P.String()}, {"comparisons": P.Array(P.WildRecord("T", {}))}, {"test": P.Fcn([P.Wildcard("D"), P.Wildcard("T")], P.Boolean())}], P.Boolean())
    errcodeBase = 32020
    def __call__(self, state, scope, pos, paramTypes, datum, operator, comparisons, test):
        if operator == "and":
            for comparison in comparisons:
                if callfcn(state, scope, test, [datum, comparison]) is False:
                    return False
            return True
        elif operator == "or":
            for comparison in comparisons:
                if callfcn(state, scope, test, [datum, comparison]) is True:
                    return True
            return False
        elif operator == "xor":
            numTrue = 0
            for comparison in comparisons:
                if callfcn(state, scope, test, [datum, comparison]) is True:
                    numTrue += 1
            return numTrue % 2 == 1
        else:
            raise PFARuntimeException("unrecognized logical operator", self.errcodeBase + 0, self.name, pos)
provide(CompoundTest())

class MissingTest(LibFcn):
    name = prefix + "missingTest"
    sig = Sig([{"datum": P.WildRecord("D", {})}, {"comparison": P.WildRecord("T", {"field": P.EnumFields("F", "D"), "operator": P.String(), "value": P.Wildcard("V")})}], P.Union([P.Null(), P.Boolean()]))
    errcodeBase = 32010
    def __call__(self, state, scope, pos, paramTypes, datum, comparison):
#         newDatumTypeFields = [{"name": x["name"], "type": removeNull(x["type"])} if x["name"] == comparison["field"] else x for x in paramTypes[0]["fields"]]
#         newParamTypes = [dict(paramTypes[0], fields=newDatumTypeFields)] + paramTypes[1:]
#         return simpleComparison(newParamTypes, datum, comparison, False, state.parser)
        out = simpleComparison(paramTypes, datum, comparison, False, state.parser, self.errcodeBase + 1, self.errcodeBase + 0, self.name, pos)
        if out is True:
            return {"boolean": True}
        elif out is False:
            return {"boolean": False}
        else:
            return out
provide(MissingTest())

class SurrogateTest(LibFcn):
    name = prefix + "surrogateTest"
    sig = Sig([{"datum": P.WildRecord("D", {})}, {"comparisons": P.Array(P.WildRecord("T", {}))}, {"missingTest": P.Fcn([P.Wildcard("D"), P.Wildcard("T")], P.Union([P.Null(), P.Boolean()]))}], P.Boolean())
    errcodeBase = 32030
    def __call__(self, state, scope, pos, paramTypes, datum, comparisons, missingTest):
        for comparison in comparisons:
            result = callfcn(state, scope, missingTest, [datum, comparison])
            if result is not None:
                if isinstance(result, dict) and list(result.keys()) == ["boolean"]:
                    return list(result.values())[0]
                else:
                    return result
        raise PFARuntimeException("no successful surrogate", self.errcodeBase + 0, self.name, pos)
provide(SurrogateTest())

class SimpleWalk(LibFcn):
    name = prefix + "simpleWalk"
    sig = Sig([{"datum": P.WildRecord("D", {})}, {"treeNode": P.WildRecord("T", {"pass": P.Union([P.WildRecord("T", {}), P.Wildcard("S")]), "fail": P.Union([P.WildRecord("T", {}), P.Wildcard("S")])})}, {"test": P.Fcn([P.WildRecord("D", {}), P.WildRecord("T", {})], P.Boolean())}], P.Wildcard("S"))
    errcodeBase = 32040
    def __call__(self, state, scope, pos, paramTypes, datum, treeNode, test):
        treeNodeTypeName = paramTypes[1]["name"]
        node = treeNode
        while True:
            if callfcn(state, scope, test, [datum, node]):
                union = node["pass"]
            else:
                union = node["fail"]
            if union is None:
                node = None
                break
            else:
                (utype, node), = list(union.items())
                if utype != treeNodeTypeName:
                    break
        return node
provide(SimpleWalk())

class MissingWalk(LibFcn):
    name = prefix + "missingWalk"
    sig = Sig([{"datum": P.WildRecord("D", {})}, {"treeNode": P.WildRecord("T", {"pass": P.Union([P.WildRecord("T", {}), P.Wildcard("S")]), "fail": P.Union([P.WildRecord("T", {}), P.Wildcard("S")]), "missing": P.Union([P.WildRecord("T", {}), P.Wildcard("S")])})}, {"test": P.Fcn([P.WildRecord("D", {}), P.WildRecord("T", {})], P.Union([P.Null(), P.Boolean()]))}], P.Wildcard("S"))
    errcodeBase = 32050
    def __call__(self, state, scope, pos, paramTypes, datum, treeNode, test):
        treeNodeTypeName = paramTypes[1]["name"]
        node = treeNode
        while True:
            result = callfcn(state, scope, test, [datum, node])
            if result is True or result == {"boolean": True}:
                union = node["pass"]
            elif result is False or result == {"boolean": False}:
                union = node["fail"]
            elif result is None:
                union = node["missing"]
            if union is None:
                node = None
                break
            else:
                (utype, node), = list(union.items())
                if utype != treeNodeTypeName:
                    break
        return node
provide(MissingWalk())

class SimpleTree(LibFcn):
    name = prefix + "simpleTree"
    sig = Sig([{"datum": P.WildRecord("D", {})}, {"treeNode": P.WildRecord("T", {"field": P.EnumFields("F", "D"), "operator": P.String(), "value": P.Wildcard("V"), "pass": P.Union([P.WildRecord("T", {}), P.Wildcard("S")]), "fail": P.Union([P.WildRecord("T", {}), P.Wildcard("S")])})}], P.Wildcard("S"))
    errcodeBase = 32060
    def __call__(self, state, scope, pos, paramTypes, datum, treeNode):
        treeNodeTypeName = paramTypes[1]["name"]
        node = treeNode
        while True:
            if simpleComparison(paramTypes, datum, node, True, state.parser, self.errcodeBase + 1, self.errcodeBase + 0, self.name, pos):
                union = node["pass"]
            else:
                union = node["fail"]
            if union is None:
                node = None
                break
            else:
                (utype, node), = list(union.items())
                if utype != treeNodeTypeName:
                    break
        return node
provide(SimpleTree())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import datetime

def pytz():
    import pytz as p
    return p

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, negativeIndex, startEnd
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn
prefix = "time."

class UTC(datetime.tzinfo):
    def utcoffset(setlf, dt):
        return datetime.timedelta(0)
    def tzname(self, dt):
        return "UTC"
    def dst(self, td):
        return datetime.timedelta(0)

def tz(dt, zone, code, name, pos):
    if zone == "":
        return dt
    else:
        try:
            return dt.astimezone(pytz().timezone(zone))
        except pytz().exceptions.UnknownTimeZoneError:
            raise PFARuntimeException("unrecognized timezone string", code, name, pos)

def tscheck(ts, code, name, pos):
    if math.isnan(ts) or ts < -62135596800 or ts > 253402300799:
        raise PFARuntimeException("timestamp out of range", code, name, pos)
    return ts

class Year(LibFcn):
    name = prefix + "year"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Int())
    errcodeBase = 40000
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        return tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos).year
provide(Year())

class MonthOfYear(LibFcn):
    name = prefix + "monthOfYear"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Int())
    errcodeBase = 40010
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        return tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos).month
provide(MonthOfYear())

class DayOfYear(LibFcn):
    name = prefix + "dayOfYear"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Int())
    errcodeBase = 40020
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        return tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos).timetuple().tm_yday
provide(DayOfYear())

class DayOfMonth(LibFcn):
    name = prefix + "dayOfMonth"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Int())
    errcodeBase = 40030
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        return tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos).day
provide(DayOfMonth())

class DayOfWeek(LibFcn):
    name = prefix + "dayOfWeek"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Int())
    errcodeBase = 40040
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        return tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos).weekday()
provide(DayOfWeek())

class HourOfDay(LibFcn):
    name = prefix + "hourOfDay"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Int())
    errcodeBase = 40050
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        return tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos).hour
provide(HourOfDay())

class MinuteOfHour(LibFcn):
    name = prefix + "minuteOfHour"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Int())
    errcodeBase = 40060
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        return tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos).minute
provide(MinuteOfHour())

class SecondOfMinute(LibFcn):
    name = prefix + "secondOfMinute"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Int())
    errcodeBase = 40070
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        return tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos).second
provide(SecondOfMinute())

class MakeTimestamp(LibFcn):
    name = prefix + "makeTimestamp"
    sig = Sig([{"year": P.Int()}, {"month": P.Int()}, {"day": P.Int()}, {"hour": P.Int()}, {"minute": P.Int()}, {"second": P.Int()}, {"millisecond": P.Int()}, {"zone": P.String()}], P.Int())
    epoch = datetime.datetime(1970, 1, 1, 0, 0, 0, 0, UTC())
    errcodeBase = 40080
    def __call__(self, state, scope, pos, paramTypes, year, month, day, hour, minute, second, millisecond, zone):
        try:
            microsecond = int(round(millisecond*1000.0))
            if zone == "":
                dt = datetime.datetime(year, month, day, hour, minute, second, microsecond, UTC())
            else:
                dt = pytz().timezone(zone).localize(datetime.datetime(year, month, day, hour, minute, second, microsecond))
        except pytz().exceptions.UnknownTimeZoneError:
            raise PFARuntimeException("unrecognized timezone string", self.errcodeBase + 0, self.name, pos)
        except ValueError:
            raise PFARuntimeException("timestamp undefined for given parameters", self.errcodeBase + 1, self.name, pos)
        else:
            return (dt.astimezone(UTC()) - self.epoch).total_seconds()
provide(MakeTimestamp())

class IsSecondOfMinute(LibFcn):
    name = prefix + "isSecondOfMinute"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}, {"low": P.Double()}, {"high": P.Double()}], P.Boolean())
    errcodeBase = 40090
    def __call__(self, state, scope, pos, paramTypes, ts, zone, low, high):
        if (low >= high):
            raise PFARuntimeException("bad time range", self.errcodeBase + 0, self.name, pos)
        second = tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 2, self.name, pos), UTC()), zone, self.errcodeBase + 1, self.name, pos).second
        return second >= low and second < high
provide(IsSecondOfMinute())

class IsMinuteOfHour(LibFcn):
    name = prefix + "isMinuteOfHour"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}, {"low": P.Double()}, {"high": P.Double()}], P.Boolean())
    errcodeBase = 40100
    def __call__(self, state, scope, pos, paramTypes, ts, zone, low, high):
        if (low >= high):
            raise PFARuntimeException("bad time range", self.errcodeBase + 0, self.name, pos)
        minute = tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 2, self.name, pos), UTC()), zone, self.errcodeBase + 1, self.name, pos).minute
        return minute >= low and minute < high
provide(IsMinuteOfHour())

class IsHourOfDay(LibFcn):
    name = prefix + "isHourOfDay"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}, {"low": P.Double()}, {"high": P.Double()}], P.Boolean())
    errcodeBase = 40110
    def __call__(self, state, scope, pos, paramTypes, ts, zone, low, high):
        if (low >= high):
            raise PFARuntimeException("bad time range", self.errcodeBase + 0, self.name, pos)
        hour = tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 2, self.name, pos), UTC()), zone, self.errcodeBase + 1, self.name, pos).hour
        return hour >= low and hour < high
provide(IsHourOfDay())

class IsDayOfWeek(LibFcn):
    name = prefix + "isDayOfWeek"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}, {"low": P.Double()}, {"high": P.Double()}], P.Boolean())
    errcodeBase = 40120
    def __call__(self, state, scope, pos, paramTypes, ts, zone, low, high):
        if (low >= high):
            raise PFARuntimeException("bad time range", self.errcodeBase + 0, self.name, pos)
        day = tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 2, self.name, pos), UTC()), zone, self.errcodeBase + 1, self.name, pos).weekday()
        return day >= low and day < high
provide(IsDayOfWeek())

class IsDayOfMonth(LibFcn):
    name = prefix + "isDayOfMonth"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}, {"low": P.Double()}, {"high": P.Double()}], P.Boolean())
    errcodeBase = 40130
    def __call__(self, state, scope, pos, paramTypes, ts, zone, low, high):
        if (low >= high):
            raise PFARuntimeException("bad time range", self.errcodeBase + 0, self.name, pos)
        day = tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 2, self.name, pos), UTC()), zone, self.errcodeBase + 1, self.name, pos).day
        return day >= low and day < high
provide(IsDayOfMonth())

class IsMonthOfYear(LibFcn):
    name = prefix + "isMonthOfYear"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}, {"low": P.Double()}, {"high": P.Double()}], P.Boolean())
    errcodeBase = 40140
    def __call__(self, state, scope, pos, paramTypes, ts, zone, low, high):
        if (low >= high):
            raise PFARuntimeException("bad time range", self.errcodeBase + 0, self.name, pos)
        month = tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 2, self.name, pos), UTC()), zone, self.errcodeBase + 1, self.name, pos).month
        return month >= low and month < high
provide(IsMonthOfYear())

class IsDayOfYear(LibFcn):
    name = prefix + "isDayOfYear"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}, {"low": P.Double()}, {"high": P.Double()}], P.Boolean())
    errcodeBase = 40150
    def __call__(self, state, scope, pos, paramTypes, ts, zone, low, high):
        if (low >= high):
            raise PFARuntimeException("bad time range", self.errcodeBase + 0, self.name, pos)
        day = tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 2, self.name, pos), UTC()), zone, self.errcodeBase + 1, self.name, pos).timetuple().tm_yday
        return day >= low and day < high
provide(IsDayOfYear())

class IsWeekend(LibFcn):
    name = prefix + "isWeekend"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Boolean())
    errcodeBase = 40160
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        day = tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos).weekday()
        return day == 5 or day == 6
provide(IsWeekend())

class IsWorkHours(LibFcn):
    name = prefix + "isWorkHours"
    sig = Sig([{"ts": P.Double()}, {"zone": P.String()}], P.Boolean())
    errcodeBase = 40170
    def __call__(self, state, scope, pos, paramTypes, ts, zone):
        dt = tz(datetime.datetime.fromtimestamp(tscheck(ts, self.errcodeBase + 1, self.name, pos), UTC()), zone, self.errcodeBase + 0, self.name, pos)
        day = dt.weekday()
        hour = dt.hour
        return (day != 5 or day != 6) and (hour >= 9 and hour < 17)
provide(IsWorkHours())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import io
import re
import urllib.request, urllib.parse, urllib.error

import titus.producer.tools as t
from titus.inspector.defs import *
from titus.util import avscToPretty

class ValidCommand(Command):
    """The 'pfa valid' command in pfainspector."""

    def __init__(self, mode):
        self.name = "valid"
        self.syntax = "valid <name>"
        self.help = "check the validity of a named PFA document\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        words = getcomplete(established)

        if len(words) == 0:
            return sorted(x + " " for x in self.mode.pfaFiles if x.startswith(active))

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)

        elif len(args) == 1 and isinstance(args[0], parser.Word):
            if args[0].text not in self.mode.pfaFiles:
                raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
            model = self.mode.pfaFiles[args[0].text]

        else:
            self.syntaxError()

        model.engine
        print("PFA document is syntactically and semantically valid")
        
class InputOutputCommand(Command):
    """The 'pfa input' and 'pfa output' commands in pfainspector."""

    def __init__(self, mode, which):
        self.name = which
        self.syntax = which + " <name> [pretty=true]"
        self.help = "view the " + which + " schema of a named PFA document\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        options = ["pretty="]
        words = getcomplete(established)

        if len(words) == 0:
            return sorted(x + " " for x in self.mode.pfaFiles if x.startswith(active))

        elif not words[-1].partial:
            return [x for x in options if x.startswith(active)]

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            options = {"pretty": True}
            while len(args) > 0 and isinstance(args[-1], parser.Option):
                opt = args.pop()
                if opt.word.text in ["pretty"]:
                    try:
                        options[opt.word.text] = opt.value.value()
                    except TypeError:
                        raise InspectorError("illegal value for {0}".format(opt.word.text))
                else:
                    raise InspectorError("option {0} unrecognized".format(opt.word.text))

            if not isinstance(options["pretty"], bool):
                raise InspectorError("pretty must be boolean")

            if len(args) == 1 and isinstance(args[0], parser.Word):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                model = self.mode.pfaFiles[args[0].text]
                node = model.obj

            else:
                self.syntaxError()

            try:
                node = node[self.name]  # "input" or "output"
            except KeyError:
                raise InspectorError("PFA document \"{0}\" is missing {1} section')".format(args[0].text, self.name))

            if isinstance(node, str):
                names = list(model.engine.parser.names.names.keys())
                if node in names:
                    node = model.engine.parser.getAvroType(node).jsonNode(set())

            if options["pretty"]:
                print(avscToPretty(node))
            else:
                t.look(node, inlineDepth=1)

class TypesCommand(Command):
    """The 'pfa types' command in pfainspector."""

    def __init__(self, mode):
        self.name = "types"
        self.syntax = "types <name> [pretty=true]"
        self.help = "list the named types defined in a named PFA document\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        options = ["pretty="]
        words = getcomplete(established)

        if len(words) == 0:
            return sorted(x + " " for x in self.mode.pfaFiles if x.startswith(active))

        elif not words[-1].partial:
            return [x for x in options if x.startswith(active)]

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            options = {"pretty": True}
            while len(args) > 0 and isinstance(args[-1], parser.Option):
                opt = args.pop()
                if opt.word.text in ["pretty"]:
                    try:
                        options[opt.word.text] = opt.value.value()
                    except TypeError:
                        raise InspectorError("illegal value for {0}".format(opt.word.text))
                else:
                    raise InspectorError("option {0} unrecognized".format(opt.word.text))

            if not isinstance(options["pretty"], bool):
                raise InspectorError("pretty must be boolean")

            if len(args) == 1 and isinstance(args[0], parser.Word):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                model = self.mode.pfaFiles[args[0].text]

            else:
                self.syntaxError()

            names = sorted(model.engine.parser.names.names.keys())
            for index, name in enumerate(names):
                node = model.engine.parser.getAvroType(name).jsonNode(set())
                print(name + ":")
                if options["pretty"]:
                    print(avscToPretty(node, 4))
                else:
                    t.look(node, inlineDepth=1)
                if index != len(names) - 1:
                    print()

            if len(names) == 0:
                print("PFA document contains no named types")
        
class UserFcnsCommand(Command):
    """The 'pfa userfcns' command in pfainspector."""

    def __init__(self, mode):
        self.name = "userfcns"
        self.syntax = "userfcns <name> [pretty=true]"
        self.help = "list details about the user functions in a named PFA document\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        options = ["pretty="]
        words = getcomplete(established)

        if len(words) == 0:
            return sorted(x + " " for x in self.mode.pfaFiles if x.startswith(active))

        elif not words[-1].partial:
            return [x for x in options if x.startswith(active)]

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            options = {"pretty": True}
            while len(args) > 0 and isinstance(args[-1], parser.Option):
                opt = args.pop()
                if opt.word.text in ["pretty"]:
                    try:
                        options[opt.word.text] = opt.value.value()
                    except TypeError:
                        raise InspectorError("illegal value for {0}".format(opt.word.text))
                else:
                    raise InspectorError("option {0} unrecognized".format(opt.word.text))

            if not isinstance(options["pretty"], bool):
                raise InspectorError("pretty must be boolean")

            if len(args) == 1 and isinstance(args[0], parser.Word):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                model = self.mode.pfaFiles[args[0].text]

            else:
                self.syntaxError()

            names = sorted(model.engineConfig.fcns)
            for index, name in enumerate(names):
                fcn = model.engineConfig.fcns[name]
                print("u." + name + ":")
                print("    parameters:")
                for pname in fcn.paramNames:
                    ptype = fcn.params[pname]
                    if options["pretty"]:
                        print("        " + pname + ": " + avscToPretty(ptype.jsonNode(set()), 10 + len(pname)).lstrip())
                    else:
                        print("        " + pname + ": " + ptype.toJson())

                if options["pretty"]:
                    print("    returns " + avscToPretty(fcn.ret.jsonNode(set()), 12).lstrip())
                else:
                    print("    returns " + fcn.ret.toJson())

                if model.engine.isRecursive("u." + name):
                    print("    recursive")
                elif model.engine.hasRecursive("u." + name):
                    print("    can call a recursive function")
                else:
                    print("    call depth: " + str(model.engine.callDepth("u." + name)))
                if model.engine.hasSideEffects("u." + name):
                    print("    can modify a cell or pool")

                calledBy = sorted(model.engine.calledBy("u." + name))
                print("    called by: " + ", ".join(calledBy) if len(calledBy) > 0 else "(none)")

                if index != len(names) - 1:
                    print()

            if len(names) == 0:
                print("PFA document contains no user functions")
        
class CallsCommand(Command):
    """The 'pfa calls' command in pfainspector."""

    def __init__(self, mode):
        self.name = "calls"
        self.syntax = "calls <name>"
        self.help = "list PFA functions called by a named PFA document\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        words = getcomplete(established)

        if len(words) == 0:
            return sorted(x + " " for x in self.mode.pfaFiles if x.startswith(active))

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)

        elif len(args) == 1 and isinstance(args[0], parser.Word):
            if args[0].text not in self.mode.pfaFiles:
                raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
            model = self.mode.pfaFiles[args[0].text]

        else:
            self.syntaxError()

        for name in sorted(model.engine.callGraph):
            print(name + ": " + ", ".join(sorted(model.engine.callGraph[name])))

class CellsPoolsCommand(Command):
    """The 'pfa cells' and 'pfa pools' commands in pfainspector."""

    def __init__(self, mode, which):
        self.name = which
        self.syntax = which + " <name> [pretty=true]"
        self.help = "list details about " + which + " in a named PFA document\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        options = ["pretty="]
        words = getcomplete(established)

        if len(words) == 0:
            return sorted(x + " " for x in self.mode.pfaFiles if x.startswith(active))

        elif not words[-1].partial:
            return [x for x in options if x.startswith(active)]

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            options = {"pretty": True}
            while len(args) > 0 and isinstance(args[-1], parser.Option):
                opt = args.pop()
                if opt.word.text in ["pretty"]:
                    try:
                        options[opt.word.text] = opt.value.value()
                    except TypeError:
                        raise InspectorError("illegal value for {0}".format(opt.word.text))
                else:
                    raise InspectorError("option {0} unrecognized".format(opt.word.text))

            if not isinstance(options["pretty"], bool):
                raise InspectorError("pretty must be boolean")

            if len(args) == 1 and isinstance(args[0], parser.Word):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                model = self.mode.pfaFiles[args[0].text]

            else:
                self.syntaxError()

            if self.name == "cells":
                node = model.engineConfig.cells
            elif self.name == "pools":
                node = model.engineConfig.pools

            names = sorted(node.keys())
            for index, name in enumerate(names):
                obj = node[name]
                if self.name == "cells":
                    preamble = "{0}: shared={1} rollback={2} type=".format(name, json.dumps(obj.shared), json.dumps(obj.rollback))
                elif self.name == "pools":
                    preamble = "{0}: shared={1} rollback={2} elements={3} type=".format(name, json.dumps(obj.shared), json.dumps(obj.rollback), len(json.loads(obj.init)))

                ptype = obj.avroType
                if options["pretty"]:
                    print(preamble + avscToPretty(ptype.jsonNode(set()), len(preamble)).lstrip())
                else:
                    print(preamble + ptype.toJson())

                if index != len(names) - 1:
                    print()

            if len(names) == 0:
                print("PFA document contains no " + self.name)
    
class ExternalizeCommand(Command):
    """The 'pfa externalize' command in pfainspector."""

    def __init__(self, mode):
        self.name = "externalize"
        self.syntax = "externalize <name> <cell-or-pool> <filename> [cell=true]"
        self.help = "turn an embedded cell or pool into an external one\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        options = ["cell="]
        words = getcomplete(established)

        if len(words) == 0:
            return sorted(x + " " for x in self.mode.pfaFiles if x.startswith(active))

        elif len(words) == 1:
            model = self.mode.pfaFiles.get(words[0].text if isinstance(words[0], parser.Word) else None)
            if model is not None:
                cells = model.obj.get("cells")
                pools = model.obj.get("pools")
                tmp = set()
                if isinstance(cells, dict):
                    tmp.update(list(cells.keys()))
                if isinstance(pools, dict):
                    tmp.update(list(pools.keys()))
                names = sorted(tmp)
            else:
                names = []
            return [x for x in names if x.startswith(active)]

        elif len(words) == 2:
            if isinstance(words[1], parser.Word) and (words[1].text + ".json").startswith(active):
                return [words[1].text + ".json"]
            else:
                return []

        elif not words[-1].partial:
            return [x for x in options if x.startswith(active)]

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            options = {"cell": True}
            while len(args) > 0 and isinstance(args[-1], parser.Option):
                opt = args.pop()
                if opt.word.text in ["cell"]:
                    try:
                        options[opt.word.text] = opt.value.value()
                    except TypeError:
                        raise InspectorError("illegal value for {0}".format(opt.word.text))
                else:
                    raise InspectorError("option {0} unrecognized".format(opt.word.text))

            if not isinstance(options["cell"], bool):
                raise InspectorError("cell must be boolean")

            if len(args) == 3 and all(isinstance(x, parser.Word) for x in args):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                model = self.mode.pfaFiles[args[0].text].obj

                cells = model.get("cells")
                pools = model.get("pools")
                obj = None
                which = ""
                if options["cell"]:
                    if isinstance(cells, dict):
                        obj = cells.get(args[1].text)
                        which = "cell"
                    if obj is None and isinstance(pools, dict):
                        obj = pools.get(args[1].text)
                        which = "pool"
                else:
                    if isinstance(pools, dict):
                        obj = pools.get(args[1].text)
                        which = "pool"
                    if obj is None and isinstance(cells, dict):
                        obj = cells.get(args[1].text)
                        which = "cell"
                if obj is None:
                    raise InspectorError("no cell or pool named {0}".format(args[1].text))

                if obj.get("source", "embedded") != "embedded":
                    raise InspectorError("{0} \"{1}\" is not currently embedded".format(which, args[1].text))

                print("Externalizing {0} \"{1}\" to {2}".format(which, args[1].text, args[2].text))
                json.dump(obj.get("init"), open(args[2].text, "w"))

                obj["source"] = "json"
                obj["init"] = args[2].text

            else:
                self.syntaxError()

class InternalizeCommand(Command):
    """The 'pfa internalize' command in pfainspector."""

    def __init__(self, mode):
        self.name = "internalize"
        self.syntax = "internalize <name> <cell-or-pool> [cell=true]"
        self.help = "turn an external cell or pool into an embedded one\n    " + self.syntax
        self.mode = mode

    def complete(self, established, active):
        """Handle tab-complete for this command's arguments.

        :type established: string
        :param established: part of the text that has been established
        :type active: string
        :param active: part of the text to be completed
        :rtype: list of strings
        :return: potential completions
        """

        options = ["cell="]
        words = getcomplete(established)

        if len(words) == 0:
            return sorted(x + " " for x in self.mode.pfaFiles if x.startswith(active))

        elif len(words) == 1:
            model = self.mode.pfaFiles.get(words[0].text if isinstance(words[0], parser.Word) else None)
            if model is not None:
                cells = model.obj.get("cells")
                pools = model.obj.get("pools")
                tmp = set()
                if isinstance(cells, dict):
                    tmp.update(list(cells.keys()))
                if isinstance(pools, dict):
                    tmp.update(list(pools.keys()))
                names = sorted(tmp)
            else:
                names = []
            return [x for x in names if x.startswith(active)]

        elif not words[-1].partial:
            return [x for x in options if x.startswith(active)]

        else:
            return []

    def action(self, args):
        """Perform the action associated with this command.

        :type args: list of titus.inspector.parser.Ast
        :param args: arguments passed to the command
        :rtype: ``None``
        :return: nothing; results must be printed to the screen
        """

        if len(args) == 1 and args[0] == parser.Word("help"):
            print(self.help)
        else:
            options = {"cell": True}
            while len(args) > 0 and isinstance(args[-1], parser.Option):
                opt = args.pop()
                if opt.word.text in ["cell"]:
                    try:
                        options[opt.word.text] = opt.value.value()
                    except TypeError:
                        raise InspectorError("illegal value for {0}".format(opt.word.text))
                else:
                    raise InspectorError("option {0} unrecognized".format(opt.word.text))

            if not isinstance(options["cell"], bool):
                raise InspectorError("cell must be boolean")

            if len(args) == 2 and all(isinstance(x, parser.Word) for x in args):
                if args[0].text not in self.mode.pfaFiles:
                    raise InspectorError("no PFA document named \"{0}\" in memory (try 'load <file> as {1}')".format(args[0].text, args[0].text))
                model = self.mode.pfaFiles[args[0].text].obj

                cells = model.get("cells")
                pools = model.get("pools")
                obj = None
                which = ""
                if options["cell"]:
                    if isinstance(cells, dict):
                        obj = cells.get(args[1].text)
                        which = "cell"
                    if obj is None and isinstance(pools, dict):
                        obj = pools.get(args[1].text)
                        which = "pool"
                else:
                    if isinstance(pools, dict):
                        obj = pools.get(args[1].text)
                        which = "pool"
                    if obj is None and isinstance(cells, dict):
                        obj = cells.get(args[1].text)
                        which = "cell"
                if obj is None:
                    raise InspectorError("no cell or pool named {0}".format(args[1].text))
        
                if obj.get("source", "embedded") != "json":
                    raise InspectorError("{0} \"{1}\" is not currently json".format(which, args[1].text))

                url = obj.get("init")
                if not isinstance(url, str):
                    raise InspectorError("{0} \"{1}\" does not have a string-valued init".format(which, args[1].text))

                print("Internalizing {0} \"{1}\" from {2}".format(which, args[1].text, url))
                if re.match("^[a-zA-Z][a-zA-Z0-9\+\-\.]*://", url) is not None:
                    init = json.load(urllib.request.urlopen(url))
                else:
                    init = json.load(open(url))

                del obj["source"]
                obj["init"] = init

            else:
                self.syntaxError()

class PFAGadget(Gadget):
    """The 'pfa' gadget in pfainspector."""

    def __init__(self, mode):
        self.commandGroup = CommandGroup("pfa", [
            ValidCommand(mode),
            InputOutputCommand(mode, "input"),
            InputOutputCommand(mode, "output"),
            TypesCommand(mode),
            UserFcnsCommand(mode),
            CallsCommand(mode),
            CellsPoolsCommand(mode, "cells"),
            CellsPoolsCommand(mode, "pools"),
            ExternalizeCommand(mode),
            InternalizeCommand(mode),
            ])
